/* 
*  wowSim: Wireless on Wheel Simulator
*
*  COPYRIGHT: 
*  Roberta Fracchia, Dario Rossi (c) 2008 
*
*  LICENSE:    
*  GPL, see LICENSE.txt for details 
*
*  DISCLAIMER: 
*  THE SOFTWARE IS PROVIDED "AS IS".  IN NO EVENT SHALL AUTHORS BE LIABLE TO ANY PARTY 
*  FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST 
*  PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION. 
* 
*                                               _____________
*                                   ..---:::::::-----------. ::::;;.
*                                .'""""""                  ;;   \  ":.
*                             .''                          ;     \   "\__.
*                           .'                            ;;      ;   \\";
*                         .'                              ;   _____;   \\/
*                       .'                               :; ;"     \ ___:'.
*                     .'--...........................    : =   ____:"    \ \
*                ..-""                               """'  o"""     ;     ; :
*           .--""  .----- ..----...    _.-    --.  ..-"     ;       ;     ; ;
*        .""_-     "--""-----'""    _-"        .-""         ;        ;    .-.
*     .'  .'                      ."         ."              ;       ;   /. |
*    /-./'                      ."          /           _..  ;       ;   ;;;|
*   :  ;-.______               /       _________==.    /_  \ ;       ;   ;;;;
*   ;  / |      """"""""""".---."""""""          :    /" ". |;       ; _; ;;;
*  /"-/  |                /   /                  /   /     ;|;      ;-" | ;';
* :-  :   """----______  /   /              ____.   .  ."'. ;;   .-"..T"   .
* '. "  ___            "":   '""""""""""""""    .   ; ;    ;; ;." ."   '--"
*  ",   __ """  ""---... :- - - - - - - - - ' '  ; ;  ;    ;;"  ."
*   /. ;  """---___                             ;  ; ;     ;|.""
*  :  ":           """----.    .-------.       ;   ; ;     ;:
*   \  '--__               \   \        \     /    | ;     ;;
*    '-..   """"---___      :   .______..\ __/..-""|  ;   ; ;
*        ""--..       """--"        m l s         .   ". . ;
*              ""------...                  ..--""      " :
*                         """"""""""""""""""    \        /
*                                                "------"
* 
*  
*  ASCII-ART: courtesy of 
*  http://www.ascii-art.de/ascii/c/car.txt
*/

/* Revision Control System LOG entries */
/* $Log: stat_mod.c,v $
 * Revision 2.1  2000/03/24 08:38:29  giaccone
 * added fflush for confidence interval
 *
 * Revision 1.1  1999/08/19 16:29:17  giaccone
 * Initial revision
 * */

/* modifica di Paolo Giaccone per evitare i problemi dovuti alla cancellazione
numerica:
lines 885, 888, 1040, 1043:
	  sqrt(varest(var_id, &st_mean[var_id]));
                 is now
	  sqrt(fabs(varest(var_id, &st_mean[var_id]))); 
          line 1241: fflush(fpout) */
/*
**  Module   : stat_mod.c 
**  Version  : 1.0
**  Author   : Andrea Francini
**  Date     : 17.03.1994
**  Synopsis : this module contains the routines dealing with the 
**             statistical aspects of a simulation run.
**	       The module "stathead.h" must be included to refer to the 
**             functions and the constants shared with the statistical 
**             interface module of the whole simulation	program.
**             All the functions, variables and constants of this module 
**	       are system-independent. 
**	       The variance of the mean value estimator is evaluated by 
**	       means of a squares accumulator.
**
**  Externally usable functions:
**	       
**	       1. void init_statistics(...)
**	       2. void d_statistics(int, double)
**	       3. void c_statistics(int, double, double)	
**	       4. void b_statistics(int, double)
**	       5. void report_statistics(FILE *, int, Booltype)
**	       6. Booltype transient_is_over(int)
**	       7. long curr_nr_of_samples(int)
**             8. double curr_mean(int)
**             9. double curr_variance(int)
**             10. double curr_accuracy(int)
**	       11. void set_ref_accuracy(double)
**	       12. void set_confid_lev(int)	
**	       13. void set_max_coll_obs(long)
**	       14. void set_min_transient_length(long)
**	       15. void set_tr_bat_size(long)
**             16. void set_tr_seq_len(long)
**	       17. void set_tr_accuracy(double)
**	       18. void set_st_bat_size(long)	
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef STAT_HDR
#include "stathead.h"
#endif


/*
** Definition of local constant values. 
*/
#define MAXNDF		(int)  30     /* Maximum number of freedom degrees   */
				      /* to use the t-Student distribution.  */

#define NR_OF_THS	(int)  10     /* Number of available accuracy        */
				      /* thresholds. 			     */

#define STAT_ARR_SIZE	(long) 2000   /* Size of the array collecting        */
				      /* the transient batch means.	     */

/*
** Statistical parameters default values. 
*/
#define DFLT_ACCURACY	(double) 0.05	
#define DFLT_NOOBS      (long) 100000
#define DFLT_TRLEN      (long)   5000
#define DFLT_TRBTSZ     (long)     30
#define DFLT_TRSQLN     (long)      3
#define DFLT_TRACC      (double) 0.99	
#define DFLT_STBTSZ     (long)  10000  /*10000*/
#define DFLT_CNFLVL     (int)       2
#define DFLT_MIN_NOOBS      (long) 0 /* PAOLO GIACCONE 12/14/01 */

/*
** Available values of confidence level. 
*/
enum   conf_level_type{p80, p90, p95, p98, p99, p999};


/*
** File pointers. 
*/
static FILE *out_fp; /* User defined file to collect partial results and     */ 
		     /* error messages.					     */


/*
** Global variables shared by the functions of this module. 
*/
static Booltype reset_called; /* Flag to check if 'var_reset(...)' has been  */
			      /* already called.			     */

static Booltype local_output; /* Flag to check the origin of the statistical */
                              /* report file.                                */
static Booltype loc_out_is_closed; /* Flag to check the status of the local  */
				   /* statistical report file.		     */

static double ref_accuracy;   /* Accuracy required to close a simulation run.*/

static int confid_lev;        /* Index to identify the required              */ 
			      /* Confidence Level. 			     */

static int no_stat_var;       /* Number of variables treated                 */
		              /* by the statistical functions.               */ 

static Booltype analysis_flag[MAX_STAT_VAR]; /* Flag to recognize the        */
					     /* variables to apply to the    */ 
					     /* complete analysis.           */

static double statdata[STAT_ARR_SIZE];       /* Array to store the Transient */
					     /* Batch Means.		     */

static long tr_bat_size;          /* Transient Batch Size.             	     */	
static long tr_seq_len;           /* Transient Batch Means Sequence Length.  */
static double tr_accuracy;        /* Transient Accuracy.                     */
static long min_transient_length; /* Minimum Transient Length.               */
static long max_coll_obs;         /* Maximum Number of Usable Observations.  */
static long min_coll_obs; /* PAOLO 14/12/01 Minimum number of observations */
static long st_bat_size;          /* Size of a Stationary Batch.             */

static double acc_bat_mean[MAX_STAT_VAR];   /* Batch Means Accumulator.      */
static double acc_sq_bat_mean[MAX_STAT_VAR];/* Accumulator for the Squares   */
					    /* of the Batch Means.	     */

static long batval_cnt[MAX_STAT_VAR];    /* Batch Values Counter.            */
static long bat_cnt[MAX_STAT_VAR];       /* Batch Counter. 	   	     */
static long tot_cnt[MAX_STAT_VAR];       /* Total Values Counter. 	     */
static double x_sum[MAX_STAT_VAR];       /* Batch Values Accumulator.        */
static double bat_time_sum[MAX_STAT_VAR];/* Batch Time Accumulator.          */
static double prev_time[MAX_STAT_VAR];   /* Previous Time Value.	     */
static double prev_value[MAX_STAT_VAR];  /* Previous Variable Value.         */

static long tr_duration[MAX_STAT_VAR];   /* Transient Duration.              */

static double accuracy_ths[NR_OF_THS];   /* Accuracy thresholds assumed as   */
					 /* reference for writing partial    */
					 /* statistical updatings.           */
static int curr_ths[MAX_STAT_VAR];  	 /* Index to identify the current    */
					 /* accuracy threshold for writing   */
					 /* statistical updatings.           */
static double conf_prcnt[NPRCNTL];	 /* Array of confidence level values.*/

static Booltype tr_end_flag[MAX_STAT_VAR]; /* Transient End Flag.            */ 
static Booltype st_end_flag[MAX_STAT_VAR]; /* Stationary Analysis End Flag.  */
static Booltype pass_end_flag;         /* Transient End Flag referred to the */
				       /* variables with analysis_flag = 0.  */ 
static Booltype st_glob_end_flag;      /* Simulation Run End Flag.           */
				      
static double delta_t;	    /* Time between the collections of consecutive   */
			    /* observations of the same variable.            */

static double st_mean[MAX_STAT_VAR];      /* Current Stationary              */
				          /* Mean values.                    */
static double st_variance[MAX_STAT_VAR];  /* Current Stationary              */
					  /* Variance values.                */
static double st_accuracy[MAX_STAT_VAR];  /* Current Stationary              */ 
					  /* Accuracy values.                */
static double st_acc_square[MAX_STAT_VAR];/* Square values Accumulator.      */

static double lower_bound[MAX_STAT_VAR];  /* Confidence Interval lower bound.*/ 
static double upper_bound[MAX_STAT_VAR];  /* Confidence Interval upper bound.*/ 

static char *proc_names[MAX_STAT_VAR];     /* Pointers to the Variables Names.*/

static double t_percentile[MAXNDF + 1][NPRCNTL];/* t-Student and Normal      */ 
static double N_percentile[NPRCNTL];            /* distributions percentiles */ 
						/* arrays.                   */


/*
** Function pointers. 
*/

/* Pointers to the functions for closing the initial transient.	       */
static void (*close_transient[MAX_STAT_VAR])(void);

/* Pointers to the functions for closing the stationary analysis.      */
static void (*close_analysis[MAX_STAT_VAR])(void);

/* Pointer to the function for closing the current simulation run.     */
static void (*close_run)(void);



/*
** Internally usable functions prototypes.
*/
static Booltype chkdelta(double, double, double);
static double varest(int, double *);
static void set_stat_array(void);
static void loc_init_stat_var(void);
static void var_reset(void);
static void preambolo(int, double, Booltype);
static void batch_method(int, double, Booltype);
static void b_preambolo(int);
static void b_batch_method(int, int);



/*
** Function: Booltype chkdelta(double a, double b, double delta)
** Remarks : this function checks if 'a' differs less then 'delta' from 'b'.
** Return  : 0 if the test answer is FALSE, 1 if the test answer is TRUE.
*/
Booltype chkdelta(double a, double b, double delta)
{
 if(b == 0.0)
	return FALSE;
 if(fabs((a - b) / b) < delta)
	return TRUE;
 else
	return FALSE;
}


/*
** Function: double varest(int var_id, double *mean) 
** Return  : the estimator of the variance of the mean value estimator.
** Remarks : the mean value estimator is also evaluated.
*/
double varest(int var_id, double *mean)
{
 double estvar;

 *mean = acc_bat_mean[var_id] / bat_cnt[var_id];
 estvar = (acc_sq_bat_mean[var_id] / bat_cnt[var_id] - (*mean) * (*mean)) / 
	  (bat_cnt[var_id] - 1.0);
 return estvar;
}


/*
** Function: void set_stat_array(void)
** Return  : None
** Remarks : this function initializes the percentiles arrays for the t-Student
**           and the standardized Normal distributions, the array 
**	     containing the sequence of thresholds dealing with the 
**	     accuracy of the mean value estimates and the array containing 
**	     the available confidence level values.
*/
void set_stat_array(void)
{
 t_percentile[1][p80] = 3.078;
 t_percentile[1][p90] = 6.314;
 t_percentile[1][p95] = 12.706;
 t_percentile[1][p98] = 31.821;
 t_percentile[1][p99] = 63.657;
 t_percentile[1][p999] = 636.619;

 t_percentile[2][p80] = 1.886;
 t_percentile[2][p90] = 2.920;
 t_percentile[2][p95] = 4.303;
 t_percentile[2][p98] = 6.965;
 t_percentile[2][p99] = 9.925;
 t_percentile[2][p999] = 31.598;

 t_percentile[3][p80] = 1.638;
 t_percentile[3][p90] = 2.353;
 t_percentile[3][p95] = 3.182;
 t_percentile[3][p98] = 4.541;
 t_percentile[3][p99] = 5.841;
 t_percentile[3][p999] = 12.941;

 t_percentile[4][p80] = 1.533;
 t_percentile[4][p90] = 2.132;
 t_percentile[4][p95] = 2.776;
 t_percentile[4][p98] = 3.747;
 t_percentile[4][p99] = 4.604;
 t_percentile[4][p999] = 8.610;

 t_percentile[5][p80] = 1.476;
 t_percentile[5][p90] = 2.015;
 t_percentile[5][p95] = 2.571;
 t_percentile[5][p98] = 3.365;
 t_percentile[5][p99] = 4.032;
 t_percentile[5][p999] = 6.859;

 t_percentile[6][p80] = 1.440;
 t_percentile[6][p90] = 1.943;
 t_percentile[6][p95] = 2.447;
 t_percentile[6][p98] = 3.143;
 t_percentile[6][p99] = 3.707;
 t_percentile[6][p999] = 5.959;

 t_percentile[7][p80] = 1.415;
 t_percentile[7][p90] = 1.895;
 t_percentile[7][p95] = 2.365;
 t_percentile[7][p98] = 2.998;
 t_percentile[7][p99] = 3.499;
 t_percentile[7][p999] = 5.405;

 t_percentile[8][p80] = 1.397;
 t_percentile[8][p90] = 1.860;
 t_percentile[8][p95] = 2.306;
 t_percentile[8][p98] = 2.896;
 t_percentile[8][p99] = 3.355;
 t_percentile[8][p999] = 5.041;

 t_percentile[9][p80] = 1.383;
 t_percentile[9][p90] = 1.833;
 t_percentile[9][p95] = 2.262;
 t_percentile[9][p98] = 2.821;
 t_percentile[9][p99] = 3.250;
 t_percentile[9][p999] = 4.781;

 t_percentile[10][p80] = 1.372;
 t_percentile[10][p90] = 1.812;
 t_percentile[10][p95] = 2.228;
 t_percentile[10][p98] = 2.764;
 t_percentile[10][p99] = 3.169;
 t_percentile[10][p999] = 4.587;

 t_percentile[11][p80] = 1.363;
 t_percentile[11][p90] = 1.796;
 t_percentile[11][p95] = 2.201;
 t_percentile[11][p98] = 2.718;
 t_percentile[11][p99] = 3.106;
 t_percentile[11][p999] = 4.437;

 t_percentile[12][p80] = 1.356;
 t_percentile[12][p90] = 1.782;
 t_percentile[12][p95] = 2.179;
 t_percentile[12][p98] = 2.681;
 t_percentile[12][p99] = 3.055;
 t_percentile[12][p999] = 4.318;

 t_percentile[13][p80] = 1.350;
 t_percentile[13][p90] = 1.771;
 t_percentile[13][p95] = 2.160;
 t_percentile[13][p98] = 2.650;
 t_percentile[13][p99] = 3.012;
 t_percentile[13][p999] = 4.221;

 t_percentile[14][p80] = 1.345;
 t_percentile[14][p90] = 1.761;
 t_percentile[14][p95] = 2.145;
 t_percentile[14][p98] = 2.624;
 t_percentile[14][p99] = 2.977;
 t_percentile[14][p999] = 4.140;

 t_percentile[15][p80] = 1.341;
 t_percentile[15][p90] = 1.753;
 t_percentile[15][p95] = 2.131;
 t_percentile[15][p98] = 2.602;
 t_percentile[15][p99] = 2.947;
 t_percentile[15][p999] = 4.073;

 t_percentile[16][p80] = 1.337;
 t_percentile[16][p90] = 1.746;
 t_percentile[16][p95] = 2.120;
 t_percentile[16][p98] = 2.583;
 t_percentile[16][p99] = 2.921;
 t_percentile[16][p999] = 4.015;

 t_percentile[17][p80] = 1.333;
 t_percentile[17][p90] = 1.740;
 t_percentile[17][p95] = 2.110;
 t_percentile[17][p98] = 2.567;
 t_percentile[17][p99] = 2.898;
 t_percentile[17][p999] = 3.965;

 t_percentile[18][p80] = 1.330;
 t_percentile[18][p90] = 1.734;
 t_percentile[18][p95] = 2.101;
 t_percentile[18][p98] = 2.552;
 t_percentile[18][p99] = 2.878;
 t_percentile[18][p999] = 3.922;

 t_percentile[19][p80] = 1.328;
 t_percentile[19][p90] = 1.729;
 t_percentile[19][p95] = 2.093;
 t_percentile[19][p98] = 2.539;
 t_percentile[19][p99] = 2.861;
 t_percentile[19][p999] = 3.883;

 t_percentile[20][p80] = 1.325;
 t_percentile[20][p90] = 1.725;
 t_percentile[20][p95] = 2.086;
 t_percentile[20][p98] = 2.528;
 t_percentile[20][p99] = 2.845;
 t_percentile[20][p999] = 3.850;

 t_percentile[21][p80] = 1.323;
 t_percentile[21][p90] = 1.721;
 t_percentile[21][p95] = 2.080;
 t_percentile[21][p98] = 2.518;
 t_percentile[21][p99] = 2.831;
 t_percentile[21][p999] = 3.819;

 t_percentile[22][p80] = 1.321;
 t_percentile[22][p90] = 1.717;
 t_percentile[22][p95] = 2.074;
 t_percentile[22][p98] = 2.508;
 t_percentile[22][p99] = 2.819;
 t_percentile[22][p999] = 3.792;

 t_percentile[23][p80] = 1.319;
 t_percentile[23][p90] = 1.714;
 t_percentile[23][p95] = 2.069;
 t_percentile[23][p98] = 2.500;
 t_percentile[23][p99] = 2.807;
 t_percentile[23][p999] = 3.767;

 t_percentile[24][p80] = 1.318;
 t_percentile[24][p90] = 1.711;
 t_percentile[24][p95] = 2.064;
 t_percentile[24][p98] = 2.492;
 t_percentile[24][p99] = 2.797;
 t_percentile[24][p999] = 3.745;

 t_percentile[25][p80] = 1.316;
 t_percentile[25][p90] = 1.708;
 t_percentile[25][p95] = 2.060;
 t_percentile[25][p98] = 2.485;
 t_percentile[25][p99] = 2.787;
 t_percentile[25][p999] = 3.725;

 t_percentile[26][p80] = 1.315;
 t_percentile[26][p90] = 1.706;
 t_percentile[26][p95] = 2.056;
 t_percentile[26][p98] = 2.479;
 t_percentile[26][p99] = 2.779;
 t_percentile[26][p999] = 3.707;

 t_percentile[27][p80] = 1.314;
 t_percentile[27][p90] = 1.703;
 t_percentile[27][p95] = 2.052;
 t_percentile[27][p98] = 2.473;
 t_percentile[27][p99] = 2.771;
 t_percentile[27][p999] = 3.690;

 t_percentile[28][p80] = 1.313;
 t_percentile[28][p90] = 1.701;
 t_percentile[28][p95] = 2.048;
 t_percentile[28][p98] = 2.467;
 t_percentile[28][p99] = 2.763;
 t_percentile[28][p999] = 3.674;

 t_percentile[29][p80] = 1.311;
 t_percentile[29][p90] = 1.699;
 t_percentile[29][p95] = 2.045;
 t_percentile[29][p98] = 2.462;
 t_percentile[29][p99] = 2.756;
 t_percentile[29][p999] = 3.659;

 t_percentile[30][p80] = 1.310;
 t_percentile[30][p90] = 1.697;
 t_percentile[30][p95] = 2.042;
 t_percentile[30][p98] = 2.457;
 t_percentile[30][p99] = 2.750;
 t_percentile[30][p999] = 3.646;

 N_percentile[p80] = 1.282;
 N_percentile[p90] = 1.645;
 N_percentile[p95] = 1.960;
 N_percentile[p98] = 2.326;
 N_percentile[p99] = 2.576;
 N_percentile[p999] = 3.300;

 accuracy_ths[0] = 1.000;
 accuracy_ths[1] = 0.500;
 accuracy_ths[2] = 0.200;
 accuracy_ths[3] = 0.100;
 accuracy_ths[4] = 0.050;
 accuracy_ths[5] = 0.020;
 accuracy_ths[6] = 0.010;
 accuracy_ths[7] = 0.005;
 accuracy_ths[8] = 0.002;
 accuracy_ths[9] = 0.001;

 conf_prcnt[p80]  = 0.8000;
 conf_prcnt[p90]  = 0.9000;
 conf_prcnt[p95]  = 0.9500;
 conf_prcnt[p98]  = 0.9800;
 conf_prcnt[p99]  = 0.9900;
 conf_prcnt[p999] = 0.9990;
 
}


/*
** Function: void loc_init_stat_var(void)
** Return  : None 
** Remarks : global initialization for the statistical variables and flags.
*/
void loc_init_stat_var(void)
{
 long i;

 for(i = 0; i < STAT_ARR_SIZE; i++)
	statdata[i] = 0.0;

 for(i = 0; i < MAX_STAT_VAR; i++)
	{
	 batval_cnt[i] = 0; 
	 bat_cnt[i] = 0;
	 tot_cnt[i] = 0;    
	 x_sum[i] = 0.0; 
	 bat_time_sum[i] = 0.0;
	 prev_time[i] = 0.0;
	 prev_value[i] = 0.0;
	 curr_ths[i] = 0;
	 tr_end_flag[i] = FALSE;
 	 st_end_flag[i] = FALSE;
	 tr_duration[i] = 0;
	 st_mean[i] = 0.0;
	 st_variance[i] = 0.0;
	 st_accuracy[i] = 0.0;
	 acc_bat_mean[i] = 0.0;
	 acc_sq_bat_mean[i] = 0.0;
	 st_acc_square[i] = 0.0;
	 lower_bound[i] = 0.0;
	 upper_bound[i] = 0.0;
	}
 reset_called = FALSE;
 pass_end_flag = FALSE;
 st_glob_end_flag = FALSE;

/*
** Set percentiles arrays.
*/
 set_stat_array();
}


/*
** Function: void var_reset(void)
** Return  : None
** Remarks : this function resets all the provisional variables, so that a 
**	     new simulation run is qualified to restart without closing
**	     the program execution (this is the case of multiple runs 
**	     obtained by changing only the initial value of the seed to 
**	     generate the pseudo-random sequences).
*/
void var_reset(void)
{
 long i;

 for(i = 0; i < STAT_ARR_SIZE; i++)
	statdata[i] = 0.0;
 for(i = 0; i < MAX_STAT_VAR; i++)
	{
	 batval_cnt[i] = 0;
	 bat_cnt[i] = 0;
	 tot_cnt[i] = 0;
	 x_sum[i] = 0.0;
	 bat_time_sum[i] = 0.0;
	 prev_time[i] = 0.0;
	 prev_value[i] = 0.0;
	 curr_ths[i] = 0;
	 if(tr_seq_len != 0)
	 	tr_end_flag[i] = FALSE;
	 else
	 	tr_end_flag[i] = TRUE;
	 st_end_flag[i] = FALSE;
	 tr_duration[i] = 0;
	 st_mean[i] = 0.0;
	 st_variance[i] = 0.0;
	 st_accuracy[i] = 0.0;
	 acc_bat_mean[i] = 0.0;
	 acc_sq_bat_mean[i] = 0.0;
	 st_acc_square[i] = 0.0;
	 lower_bound[i] = 0.0;
	 upper_bound[i] = 0.0;
	}
 reset_called = TRUE;
 pass_end_flag = FALSE;
 st_glob_end_flag = FALSE;
}
	 

/*
** Function: void init_statistics(...)
** Return  : None
** Remarks : this function sets the variables used by the statistical 
**           routines, according to the programmer (or user) needs.
**
** Parameters description:
**
**	     1. n_stat_var      : number of variables to apply to the mean and 
**		variance estimation;             
**           2. significant_var : array of boolean flags to recognize the 
**		variables to apply to the complete statistical analysis (if 
**		NULL the complete analysis applies to all the variables);
**           3. var_names       : array of string pointers; the items referred 
**		to by the pointers are the names of the observed variables 
**		(NULL if automatic names assignement is required);
**           4. end_transient   : array of pointers to the functions to execute 
**		the operations required when the i-th variable transient is 
**		over (NULL if no operations are required);
**           5. end_analysis    : array of pointers to the functions to execute 
**		the operations required when the i-th variable stationary 
**              accuracy is achieved (NULL if no operations are required);
**           6. end_sim_run     : pointer to the function to execute the 
**		operations  required when the simulation run is over (Fatal 
**		Error if NULL); 
**           7. stat_file_p     : pointer to the output file to write onto 
**		the error messages and the statistical report (if NULL, the 
**		file name assignment becomes automatic);
*/
void init_statistics(
		     int n_stat_var,
		     Booltype significant_var[], 
		     char *var_names[],
                     void (*end_transient[])(), 
		     void (*end_analysis[])(),
	             void (*end_sim_run)(), 
	             FILE *stat_file_p
	            )
{
 int i;

/*		           
** Report file management.
*/
 if(stat_file_p == NULL)
	{
	 out_fp = fopen("debug/std_stat.rep", "w");
	 fprintf(out_fp, "\n# This is the standard file to collect ");
	 fprintf(out_fp, "the informations about the statistical\n");
	 fprintf(out_fp, "# behaviour of the simulation runs.\n\n");
         local_output = TRUE;
 	 loc_out_is_closed = FALSE;
	}
 else
	{
	 out_fp = stat_file_p;
         local_output = FALSE;
	}

/*
** 'no_stat_var' assignement.
*/
 if((n_stat_var < 1)||(n_stat_var > MAX_STAT_VAR))
	{
	 fprintf(out_fp, "\n# Error on statistical variables number ");
         fprintf(out_fp, "selection!\n");
	 printf( "\n# Error on statistical variables number ");
         printf( "selection!\n");
	 if(local_output == TRUE)
	 	fclose(out_fp);	 
	 exit(1);
	}
 else
	no_stat_var = n_stat_var;

/*
** Checking the pointer to the function to close the simulation run.
*/
 if(end_sim_run == NULL)
	{
 	 fprintf(out_fp, "\n# Fatal Error: the simulation run ");
	 fprintf(out_fp, "can't be closed!\n");
 	 printf( "\n# Fatal Error: the simulation run ");
	 printf( "can't be closed!\n");
	 if(local_output == TRUE)
		fclose(out_fp);
	 exit(1);
	}
 else
	close_run = end_sim_run;

/*
** 'analysis_flag[]' assignement.
*/
 if(significant_var == NULL)
	for(i = 0; i < no_stat_var; i++)
		analysis_flag[i] = TRUE;	
 else
	for(i = 0; i < no_stat_var; i++)
		analysis_flag[i] = significant_var[i];		

/*
** Default values assignment for the set of statistical parameters.
*/
 ref_accuracy = DFLT_ACCURACY;
 min_transient_length = DFLT_TRLEN;
 max_coll_obs = DFLT_NOOBS;
 min_coll_obs = DFLT_MIN_NOOBS; /* PAOLO GIACCONE 12/14/01 */
 tr_bat_size = DFLT_TRBTSZ;
 tr_seq_len = DFLT_TRSQLN;
 tr_accuracy = DFLT_TRACC;
 st_bat_size = DFLT_STBTSZ;
 confid_lev = DFLT_CNFLVL;

/*
** Variables names management. 
*/
 if(var_names != NULL)
	for(i = 0; i < no_stat_var; i++)
		proc_names[i] = var_names[i];
 else
	for(i = 0; i < no_stat_var; i++)
		{
	         proc_names[i] = (char *)(malloc(sizeof(char[10])));
	         if(proc_names[i] == NULL)
		    {
		     fprintf(out_fp, "\n# Error on memory allocation!");
	             printf( "\n# Error on memory allocation in init_statistics()!");
		     exit(1);
		    }
		 sprintf(proc_names[i], "var[%d]", i);
	        }

/*
** Closing transient functions pointers assignement.
*/
 if(end_transient != NULL)
	 for(i = 0; i < no_stat_var; i++)
 	 	close_transient[i] = end_transient[i];
 else
	 for(i = 0; i < no_stat_var; i++)
 	 	close_transient[i] = NULL;

/*
** Single analysis closing functions pointers assignement.
*/
 if(end_analysis != NULL)
	 for(i = 0; i < no_stat_var; i++)
 	 	close_analysis[i] = end_analysis[i];
 else
	 for(i = 0; i < no_stat_var; i++)
 	 	close_analysis[i] = NULL;

/*
** Local Statistical Variables Initialization.
*/
 loc_init_stat_var();
}	  


/*
** Function : void preambolo(int var_id, double var_val, Booltype is_discrete)
** Return   : None
** Remarks  : this function implements a rule of thumb to detect the 
**            initial transient.
**	      is_discrete = TRUE  if a discrete-time  variable is analyzed;
**	      is_discrete = FALSE if a continous-time variable is analyzed.
** The Rule : in a time series of batch means xmean1(m0), xmean2(m0), ...
**            the initial transient is over after b0 batches, i.e., after 
**            n0 = b0 * m0 observations, if the k most recent batch means
**	      all fall within an interval of relative half-width 'delta'.
*/
void preambolo(int var_id, double var_val, Booltype is_discrete)
{
 long i;
 int j;

if(analysis_flag[var_id] == FALSE)	/* Only data referring to interesting */
	{				/* variables are processed by the     */
	 tot_cnt[var_id]++;		/* transient detection rule.          */
	 return;
	}
 batval_cnt[var_id]++;
//fprintf(out_fp,"var_id:%d varval:%f tot_cnt:%d batval:%d\n",var_id,var_val,tot_cnt[var_id],batval_cnt[var_id]);
//printf("var_id:%d varval:%f tot_cnt:%d batval:%d\n",var_id,var_val,tot_cnt[var_id],batval_cnt[var_id]);
 if(is_discrete == TRUE)
 	x_sum[var_id] += var_val;
 else
	{
	 x_sum[var_id] += var_val * delta_t;
	 bat_time_sum[var_id] += delta_t;
	}
 if(batval_cnt[var_id] == tr_bat_size) 
    {
     if(is_discrete == TRUE)
     	statdata[(bat_cnt[var_id] % tr_seq_len) * no_stat_var + var_id] 
			= x_sum[var_id] / tr_bat_size;
     else
	statdata[(bat_cnt[var_id] % tr_seq_len) * no_stat_var + var_id] 
			= x_sum[var_id] / bat_time_sum[var_id];
     tot_cnt[var_id] += tr_bat_size;
     batval_cnt[var_id] = 0;
     x_sum[var_id] = 0.0;	
     bat_time_sum[var_id] = 0.0;
     bat_cnt[var_id]++;
     //printf("batcnt:%d (tr_seq_len:%d\n",bat_cnt[var_id],tr_seq_len);
     
     if(bat_cnt[var_id] >= tr_seq_len)
	{
	 double mean_ov_mean;

	 Booltype again = TRUE;
	 double accmean = 0.0;

	 for(i = 0; i < tr_seq_len; i++)
	 	accmean += statdata[i * no_stat_var + var_id];
	 mean_ov_mean = accmean / tr_seq_len;
         //printf("mean:%f  stat:%f delta:%f\n",mean_ov_mean,statdata[var_id], tr_accuracy);
/*       
** Checking the initial transient closing condition.
*/
	 for(i = 0; (i < tr_seq_len) && (again == TRUE); i++)
	     again = chkdelta(statdata[i * no_stat_var + var_id], 
			      mean_ov_mean, tr_accuracy);  
         if((again == TRUE) && (tot_cnt[var_id] >= min_transient_length))
	     {
	      tr_duration[var_id] = tot_cnt[var_id];
	      report_statistics(out_fp, var_id, TRUE);
	      tr_end_flag[var_id] = TRUE;
/*
** Closing the initial transient for variables with 'analysis_flag == FALSE'.
** These transients are over when the first interesting transient is over.
*/
	      if(pass_end_flag == FALSE)
		{
		 pass_end_flag = TRUE;
	         for(j = 0; j < no_stat_var; j++)
		    if(analysis_flag[j] == FALSE) 
			{
			 tr_end_flag[j] = TRUE;
			 tr_duration[j] = tot_cnt[j];
			 report_statistics(out_fp, j, TRUE);
	      		 batval_cnt[j] = 0;
	      	  	 bat_cnt[j] = 0;
	           	 tot_cnt[j] = 0;
	      		 x_sum[j] = 0.0;
			 bat_time_sum[j] = 0.0;
	      		 if(close_transient[j] != NULL)	
				(*close_transient[j])();
			}
		}      		
/*
** Refreshing the variables employed during the stationary analysis.
*/
	      bat_cnt[var_id] = 0;
	      tot_cnt[var_id] = 0;
/*
** Refreshing the system-dependent variables to be reinitialized.
*/
	      if(close_transient[var_id] != NULL)
	      	     (*close_transient[var_id])();	
	     } /* if(again == TRUE) */
        } /* if(bat_cnt ...) */
    } /* if(batval_cnt ...) */
} /* function */


/*
** Function  : void batch_method(int var_id, double var_val, 
**				 Booltype is_discrete)
** Return    : None
** Remarks   : this function implements the 'batch means method' of 
**             stationary data collection and analysis.
**	       is_discrete = TRUE  if a discrete-time  variable is analyzed;
**	       is_discrete = FALSE if a continous-time variable is analyzed.
** The Method: the recorded sequence of original observations is divided 
**	       into a series of nonoverlapping batches of size 'm', and the 
**             respective batch means are used as (secondary) output data in
**             statistical analysis of the simulation results.
*/
void batch_method(int var_id, double var_val, Booltype is_discrete)
{
 long kb;

 batval_cnt[var_id]++;
 if(is_discrete == TRUE)
	{
 	 x_sum[var_id] += var_val;
	 st_acc_square[var_id] += var_val*var_val;
	}
 else
	{
	 x_sum[var_id] += var_val * delta_t;
	 bat_time_sum[var_id] += delta_t;
	}	 
 //printf("batval_cnt[%d]:%d bat_size:%d\n",var_id,batval_cnt[var_id],st_bat_size);
 if(batval_cnt[var_id] == st_bat_size) 
    {
     double new_bat_mean;

     if(is_discrete == TRUE)  
   	new_bat_mean = x_sum[var_id] / st_bat_size;
     else
	{
	 new_bat_mean = x_sum[var_id] / bat_time_sum[var_id];
	 st_acc_square[var_id] += new_bat_mean*new_bat_mean;
	}
     acc_bat_mean[var_id] += new_bat_mean;
     acc_sq_bat_mean[var_id] += new_bat_mean * new_bat_mean;
     tot_cnt[var_id] += st_bat_size;
     batval_cnt[var_id] = 0;
     x_sum[var_id] = 0.0;	
     bat_time_sum[var_id] = 0.0;
     bat_cnt[var_id]++;
     //printf("bat_cnt[%d]:%d\n",var_id, bat_cnt[var_id]);
     if((kb = bat_cnt[var_id]) >= 2)
	{
	 if(kb <= MAXNDF + 1)
	   st_accuracy[var_id] = t_percentile[kb - 1][confid_lev] *
	     		sqrt(fabs(varest(var_id, &st_mean[var_id])));
	 else
	   st_accuracy[var_id] = N_percentile[confid_lev] *
	     		sqrt(fabs(varest(var_id, &st_mean[var_id])));

	 lower_bound[var_id] = st_mean[var_id] - st_accuracy[var_id];
	 upper_bound[var_id] = st_mean[var_id] + st_accuracy[var_id];

	 if(st_mean[var_id] != 0.0)
	   	st_accuracy[var_id] /= st_mean[var_id];
	 else
	   	st_accuracy[var_id] = 100.0;

	 if(is_discrete == TRUE)
	   st_variance[var_id] = (st_acc_square[var_id] - tot_cnt[var_id] * 
				 st_mean[var_id] * st_mean[var_id]) / 
				 (tot_cnt[var_id] - 1.0);
	 else
           st_variance[var_id] = (st_acc_square[var_id] - kb * 
				 st_mean[var_id]*st_mean[var_id]) / (kb - 1.0);

	 if((st_accuracy[var_id] < accuracy_ths[curr_ths[var_id]]) &&
	    (curr_ths[var_id] < NR_OF_THS))
         	{
		 report_statistics(out_fp, var_id, FALSE);
		 curr_ths[var_id]++;
	 	}

	 if(analysis_flag[var_id] == TRUE)
         { /* PAOLO GIACCONE - added new condition to verity min_coll_obs */

	   if(tot_cnt[var_id]>=min_coll_obs) {
	     if((st_accuracy[var_id] < ref_accuracy) || 
		(tot_cnt[var_id] >= max_coll_obs))
	    	{
		 int j;
	     	 Booltype again = TRUE;

	     	 st_end_flag[var_id] = TRUE;
	     	 if(close_analysis[var_id] != NULL)
			   (*close_analysis[var_id])();
	      	 for(j = 0; (j < no_stat_var) && (again == TRUE); j++)
			again = st_end_flag[j] || (analysis_flag[j] == FALSE);
	     	 st_glob_end_flag = again;
	     	 if(st_glob_end_flag == TRUE)
			{
		 	 for(j = 0; j < no_stat_var; j++)
				report_statistics(out_fp, j, FALSE);
		 	 if(local_output == TRUE)
				{
				 fclose(out_fp);
				 loc_out_is_closed = TRUE;
				}
		 	 close_run();	
		 	 var_reset();
			} /* if(st_glob_end_flag ...) */
		} /* if(st_accuracy[var_id]...) */
               }/* if(tot_cnt[var_id]>=min_coll_obs) { */ 
	     } /* if(analysis_flag[var_id] ...) */
        } /* if((kb = ... ) */
    } /* if(batval ... ) */
} /* function */


/*
** Function  : void b_preambolo(int var_id)
** Return    : None
** Remarks   : this function manages the initial transient for binary 
**	       discrete-time variables; if a significant binary variable is 
**	       observed, the initial transient is over when 
**	       'min_transient_length' observations have been collected (the 
**	       transient detection rule of 'preambolo' is not applied here).
*/
void b_preambolo(int var_id)
{
 int j;

 tot_cnt[var_id]++;
 if((analysis_flag[var_id] == TRUE) && 
    (tot_cnt[var_id] >= min_transient_length))
	{
	 tr_duration[var_id] = tot_cnt[var_id];
	 report_statistics(out_fp, var_id, TRUE);
	 tr_end_flag[var_id] = TRUE;
         tot_cnt[var_id] = 0;
/*
** Closing the initial transient for variables with 'analysis_flag == FALSE'.
** These transients are over when the first interesting transient is over.
*/
	 if(pass_end_flag == FALSE)
		{
		 pass_end_flag = TRUE;
	         for(j = 0; j < no_stat_var; j++)
		    if(analysis_flag[j] == FALSE) 
			{
			 tr_end_flag[j] = TRUE;
			 tr_duration[j] = tot_cnt[j];
			 report_statistics(out_fp, j, TRUE);
	      		 batval_cnt[j] = 0;
	      	  	 bat_cnt[j] = 0;
	           	 tot_cnt[j] = 0;
	      		 x_sum[j] = 0.0;
			 bat_time_sum[j] = 0.0;
	      		 if(close_transient[j] != NULL)	
				(*close_transient[j])();
			}
		}      		
/*
** Refreshing the system-dependent variables to be reinitialized.
*/
	 if(close_transient[var_id] != NULL)
	 	(*close_transient[var_id])();	
	} /* if(analysis_flag ...) */
} /* function */
   

/*
** Function  : void b_batch_method(int var_id, int var_val)
** Return    : None
** Remarks   : this function implements the 'batch means method' of 
**             stationary data collection and analysis, for binary 
**	       discrete-time variables.
**	       For significant binary variables the analysis is over when 
**	       the number of collected observations exceeds the threshold 
**	       given by 'max_coll_obs' (no tests over the achieved accuracy
**	       are carried out).
** The Method: the recorded sequence of original observations is divided 
**	       into a series of non-overlapping batches of size 'm', and the 
**             respective batch means are used as (secondary) output data in
**             the statistical analysis of the simulation results.
*/
void b_batch_method(int var_id, int var_val)
{
 long kb;

 batval_cnt[var_id]++;
 x_sum[var_id] += var_val;
 st_acc_square[var_id] += var_val*var_val;
 if(batval_cnt[var_id] == st_bat_size) 
    {
     double new_bat_mean;

     new_bat_mean = x_sum[var_id] / st_bat_size;
     acc_bat_mean[var_id] += new_bat_mean;
     acc_sq_bat_mean[var_id] += new_bat_mean * new_bat_mean;
     tot_cnt[var_id] += st_bat_size;
     batval_cnt[var_id] = 0;
     x_sum[var_id] = 0.0;	
     bat_cnt[var_id]++;
     if((kb = bat_cnt[var_id]) >= 2)
	{
	 if(kb <= MAXNDF + 1)
	   st_accuracy[var_id] = t_percentile[kb - 1][confid_lev] *
	     		sqrt(fabs(varest(var_id, &st_mean[var_id])));
	 else
	   st_accuracy[var_id] = N_percentile[confid_lev] *
	     		sqrt(fabs(varest(var_id, &st_mean[var_id])));

	 lower_bound[var_id] = st_mean[var_id] - st_accuracy[var_id];
	 if(lower_bound[var_id] < 0.0)
		lower_bound[var_id] = 0.0;
	 upper_bound[var_id] = st_mean[var_id] + st_accuracy[var_id];
	 if(upper_bound[var_id] > 1.0)
		upper_bound[var_id] = 1.0;
  
	 if((st_mean[var_id] != 0.0) && (st_accuracy[var_id] != 0.0))
	 	st_accuracy[var_id] /= st_mean[var_id];
	 else
		{
		 st_accuracy[var_id] = 100.0;
		 if(st_mean[var_id] == 0.0)
	   		{
	    		 lower_bound[var_id] = 0.0;
	    		 upper_bound[var_id] = 1.0 - 
				pow(1.0 - conf_prcnt[confid_lev], 
				1.0 / tot_cnt[var_id]);
			}
		}

	 st_variance[var_id] = (st_acc_square[var_id] - tot_cnt[var_id] *
			        st_mean[var_id]*st_mean[var_id]) / 
				(tot_cnt[var_id] - 1.0);

	 if((st_accuracy[var_id] < accuracy_ths[curr_ths[var_id]]) &&
	    (curr_ths[var_id] < NR_OF_THS))
         	{
		 report_statistics(out_fp, var_id, FALSE);
		 curr_ths[var_id]++;
	 	}

	 if(analysis_flag[var_id] == TRUE)
	    {
	     if(tot_cnt[var_id] >= max_coll_obs)
	    	{
	     	 int j;
	     	 Booltype again = TRUE;

	     	 st_end_flag[var_id] = TRUE;
	     	 if(close_analysis[var_id] != NULL)
			   (*close_analysis[var_id])();
	      	 for(j = 0; (j < no_stat_var) && (again == TRUE); j++)
			again = st_end_flag[j] || (analysis_flag[j] == FALSE);
	     	 st_glob_end_flag = again;
	     	 if(st_glob_end_flag == TRUE)
			{
		 	 for(j = 0; j < no_stat_var; j++)
				report_statistics(out_fp, j, FALSE);
		 	 if(local_output == TRUE)
				{
				 fclose(out_fp);
				 loc_out_is_closed = TRUE;
				}
		 	 close_run();	
		 	 var_reset();
			} /* if(st_glob_end_flag ...) */
		} /* if(st_accuracy[var_id]...) */
	     } /* if(analysis_flag[var_id] ...) */
        } /* if((kb = ... ) */
    } /* if(batval ... ) */
} /* function */


/*
** Function : void d_statistics(int var_id, double var_val)
** Return   : None
** Remarks  : externally usable function to lead the collection and analysis
**            of discrete-time statistical variables.
**
** Parameters description:
**
**	      1. var_id : variable identifier;
**            2. var_val: variable value.
*/
void d_statistics(int var_id, double var_val)
{
//fprintf(out_fp,"var_id:%d val:%f flag:%d\n",var_id, var_val,tr_end_flag[var_id]);
 /* if (var_id==1)
     printf("var_id:%d val:%f flag:%d\n",var_id, var_val,tr_end_flag[var_id]);
*/ if(tr_end_flag[var_id] == FALSE)
 	preambolo(var_id, var_val, TRUE);
 else
   	batch_method(var_id, var_val, TRUE);
}


/*
** Function : void c_statistics(int var_id, double var_val, double sim_time)
** Return   : None
** Remarks  : externally usable function to lead the collection and analysis
**            of continous-time statistical variables.
**
** Parameters description:
**
**	      1. var_id   : variable identifier;
**            2. var_val  : variable value;
**            3. sim_time : current simulation time.
*/
void c_statistics(int var_id, double var_val, double sim_time)
{
 delta_t = sim_time - prev_time[var_id];
 if(tr_end_flag[var_id] == FALSE)
	preambolo(var_id, prev_value[var_id], FALSE);
 else
	batch_method(var_id, prev_value[var_id], FALSE);
 if(reset_called == FALSE)
	{
 	 prev_time[var_id] = sim_time;
	 prev_value[var_id] = var_val;
	}
 else
	reset_called = FALSE;
}


/*
** Function : void b_statistics(int var_id, int var_val)
** Return   : None
** Remarks  : externally usable function to lead the collection and analysis
**            of binary statistical variables.
**
** Parameters description:
**
**	      1. var_id : variable identifier;
**            2. var_val: variable value (0 or 1).
*/
void b_statistics(int var_id, int var_val)
{
 if(tr_end_flag[var_id] == FALSE)
	b_preambolo(var_id);
 else
   	b_batch_method(var_id, var_val);
}


/*
** Function: void report_statistics(FILE *fpout, int var_id, 
**				    Booltype still_in_transient) 
** Return  : None
** Remarks : this function writes onto the file pointed by 'fpout' 
**           the statistical report referring to the variable indexed 
**           by 'var_id'.
**           
**           still_in_transient = TRUE : end of transient report is required;
**           still_in_transient = FALSE: stationary report is required;
*/
void report_statistics(FILE *fpout, int var_id, Booltype still_in_transient)
{
 static Booltype rep_intro = FALSE;

 if((local_output == TRUE) && (loc_out_is_closed == TRUE))
	{
	 if((out_fp = fopen("std_stat.rep", "a")) != NULL)
		fpout = out_fp;
	 else
		{
		 out_fp = fopen("std_stat.rep", "w");
		 fpout = out_fp;
		}
	 loc_out_is_closed = FALSE;
	}
	 
 if(rep_intro == FALSE)
	{
	 fprintf(fpout, "\n# The Confidence Level is %G%% ", 
			 100.0 * conf_prcnt[confid_lev]);
	 fprintf(fpout, "for all the observed variables.\n\n\n");
	 rep_intro = TRUE;
	}

 if(still_in_transient == TRUE)
	{
	 fprintf(fpout, "\n# Transient duration %s : %ld",
		 proc_names[var_id], tr_duration[var_id]);
	 if(analysis_flag[var_id] == TRUE)
		 fprintf(fpout, " (significant variable)");
	 fprintf(fpout, "\n\n");
	}
 else
	{
	 fprintf(fpout, "\n# Variable : %s ", proc_names[var_id]);
	 if(analysis_flag[var_id] == TRUE)
		 fprintf(fpout, " (significant variable)");
	 fprintf(fpout, "\n\n# Nsamples : %ld", tot_cnt[var_id]);
	 fprintf(fpout, "\n# Mean     : %G", st_mean[var_id]);
	 fprintf(fpout, "\n# Variance : %G", st_variance[var_id]);
	 if(st_accuracy[var_id] != 100.0)
	 	fprintf(fpout, "\n# Accuracy : %G", st_accuracy[var_id]);
	 else
		fprintf(fpout, "\n# Accuracy is not available !!");
	 fprintf(fpout, "\n# Confidence Interval Lower Bound : %G", 
			lower_bound[var_id]);	 
	 fprintf(fpout, "\n# Confidence Interval Upper Bound : %G\n\n", 
			upper_bound[var_id]);
         fflush(fpout);
        }
}


/*
** Function: Booltype transient_is_over(int var_id)
** Return  : TRUE  if the initial transient is over for the variable indexed 
**	     	   by 'var_id';
**	     FALSE if the initial transient is not over for the variable 
**	     	   indexed by 'var_id'.
*/
Booltype transient_is_over(int var_id)
{
 return tr_end_flag[var_id];
}


/*
** Function: long curr_nr_of_samples(int var_id)
** Return  : the number of observations collected during the current 
**	     simulation phase (transient or stationary state) for the 
**	     variable indexed by 'var_id'.
*/
long curr_nr_of_samples(int var_id)
{
 return tot_cnt[var_id];
}


/*
** Function: double curr_mean(int var_id)
** Return  : the current stationary mean value referring to the variable 
**           indexed by 'var_id'.
** Remarks : the returned value is the most recently evaluated available 
**           value.
*/
double curr_mean(int var_id)
{
 return st_mean[var_id];
}


/*
** Function: double curr_variance(int var_id)
** Return  : the current stationary variance value referring to the variable 
**           indexed by 'var_id'.
** Remarks : the returned value is the most recently evaluated available 
**           value.
*/
double curr_variance(int var_id)
{
 return st_variance[var_id];
}


/*
** Function: double curr_accuracy(int var_id)
** Return  : the current stationary accuracy value referring to the variable 
**           indexed by 'var_id'.
** Remarks : the returned value is the most recently evaluated available 
**           value.
*/
double curr_accuracy(int var_id)
{
 return st_accuracy[var_id];
}


/*
** Function: void set_ref_accuracy(double value)
** Return  : None
** Remarks : this function may be called to set 'ref_accuracy' value.
**
**	     'value' is the new 'ref_accuracy' value.
**	     Legal range of 'value' : (0.0,1.0) .
**
**           Default value of 'ref_accuracy' :  0.05 (DFLT_ACCURACY)
**
**           IF 	(value = DOUBLE_UNCHANGED)
**	     THEN 	the default value is assigned to 'ref_accuracy'
**	     ENDIF
** 	
**           IF         (value is out of range)
**           THEN       an error message appears into the proper file and 
**                      the default value is assigned to 'ref_accuracy'
**           ELSE	'value' is assigned to 'ref_accuracy'
**	     ENDIF
**
**
**	     WARNING !! - The call to this function must mandatorily 
**			- follow the call to 'init_statistics(...)'.
**	    
**	     
*/
void set_ref_accuracy(double value)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(ref_accuracy != DFLT_ACCURACY)
	{
	 printf("\nFatal Error on using 'set_ref_accuracy(...)' !!\n");
         printf("\nAccuracy level setting is not permitted ");
         printf("before calling \n\n\t'init_statistics(...)' !!\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

 if(value == DOUBLE_UNCHANGED)
	 return;
 if((value <= 0.0) || (value >= 1.0))
	{
	 fprintf(out_fp, "\n# Warning: your Accuracy was out of range!");	
	 fprintf(out_fp, "\n#          Accuracy is now %f.\n", ref_accuracy);	
	 return;
	}
 ref_accuracy = value;
} 


/*
** Function: void set_confid_lev(int index)
** Return  : None
** Remarks : this function may be called to set 'confid_lev' value.
**
**	     'index' is the new 'confid_lev' value.
**	     Legal range of 'index' : [0,NPRCNTL-1] .
**
**           Default value of 'confid_lev' :  2 (DFLT_CNFLVL) ( => 95%) 
**
**           IF 	(index = INT_UNCHANGED)
**	     THEN 	the default value is assigned to 'confid_lev'
**	     ENDIF
** 	
**           IF         (index is out of range)
**           THEN       an error message appears into the proper file and 
**                      the default value is assigned to 'confid_lev'
**           ELSE	'index' is assigned to 'confid_lev'
**	     ENDIF
**
**
**	     WARNING !! - The call to this function must mandatorily 
**			- follow the call to 'init_statistics(...)'.
**	     
*/
void set_confid_lev(int index)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(confid_lev != DFLT_CNFLVL)
	{
	 printf("\nFatal Error on using 'set_confid_lev(...)' !!\n");
         printf("\nConfidence Level setting is not permitted ");
         printf("before calling \n\n\t'init_statistics(...)' !!\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

 if(index == INT_UNCHANGED)
	 return;
 if((index < 0) || (index >= NPRCNTL))
	{
	 fprintf(out_fp, "\n# Warning: value mysmatch on Confidence Level ");
         fprintf(out_fp, "selection!");
         fprintf(out_fp, "\n#          Confidence Level is now %G%%.\n", 
		          100.0 * conf_prcnt[confid_lev]);
	 return;
	}
 confid_lev = index;
} 


/*
** Function: void set_max_coll_obs(long noobs)
** Return  : None
** Remarks : this function may be called to set 'max_coll_obs' value.
**
**	     'noobs' is the new 'max_coll_obs' value.
**	     Legal range of 'noobs' : noobs > 0 .
**
**           Default value of 'max_coll_obs' :  100000 (DFLT_NOOBS)
**
**           IF 	(noobs = LONG_UNCHANGED)
**	     THEN 	the default value is assigned to 'max_coll_obs'
**	     ENDIF
** 	
**           IF         (noobs is out of range)
**           THEN       an error message appears into the proper file and 
**                      the default value is assigned to 'max_coll_obs'
**           ELSE	'noobs' is assigned to 'max_coll_obs'
**	     ENDIF
**
**
**	     WARNING !! - The call to this function must mandatorily 
**			- follow the call to 'init_statistics(...)'.
**	     
*/
void set_max_coll_obs(long noobs)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(max_coll_obs != DFLT_NOOBS)
	{
	 printf("\nFatal Error on using 'set_max_coll_obs(...)' !!\n");
         printf("\nMaximum Number of Usable observations setting is not ");
         printf("permitted before calling \n\n\t'init_statistics(...)' !!\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

 if(noobs == LONG_UNCHANGED)
	 return;
 if(noobs <= 0) 
	{
	 fprintf(out_fp, "\n# Warning: input error on Maximum Number of ");
	 fprintf(out_fp, "Usable Observations selection!"); 	
         fprintf(out_fp, "\n#          Maximum Number of Usable Observations ");
	 fprintf(out_fp, "is now %ld.\n", max_coll_obs);
	 return;
	}
 max_coll_obs = noobs;
} 

/* This function has been added by PAOLO GIACCONE 12/14/01 */
void set_min_coll_obs(long noobs)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(min_coll_obs != DFLT_MIN_NOOBS)
	{
	 printf("\nFatal Error on using 'set_min_coll_obs(...)' !!\n");
         printf("\nMaximum Number of Usable observations setting is not ");
         printf("permitted before calling \n\n\t'init_statistics(...)' !!\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

/* if(noobs == LONG_UNCHANGED)
   return; */
 if(noobs < 0) 
	{
	 fprintf(out_fp, "\n# Warning: input error on Maximum Number of ");
	 fprintf(out_fp, "Usable Observations selection!"); 	
         fprintf(out_fp, "\n#          Maximum Number of Usable Observations ");
	 fprintf(out_fp, "is now %ld.\n", min_coll_obs);
	 return;
	}
 min_coll_obs = noobs;
} 


/*
** Function: void set_min_transient_length(long value)
** Return  : None
** Remarks : this function may be called to set 'min_transient_length' value.
**
**	     'value' is the new 'min_transient_length' value.
**	     Legal range of 'value' : value >= 0 .
**
**           Default value of 'value' :  5000 (DFLT_NOOBS)
**
**           IF 	(value = LONG_UNCHANGED)
**	     THEN 	the default value is assigned to 
**                      'min_transient_length'. 
**	     ENDIF
** 	
**           IF         (value is out of range)
**           THEN       an error message appears into the proper file and 
**                      the default value is assigned to 
**		        'min_transient_length' 
**           ELSE	'value' is assigned to 'min_transient_length'
**	     ENDIF
**
**
**	     WARNING !! - The call to this function must mandatorily 
**			- follow the call to 'init_statistics(...)'.
**	     
*/
void set_min_transient_length(long value)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(min_transient_length != DFLT_TRLEN)
	{
	 printf("\nFatal Error on using 'set_min_transient_length(...)' !!\n");
         printf("\nMinimum Transient Length setting is not permitted ");
         printf("before calling \n\n\t'init_statistics(...)'.\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

 if(value == LONG_UNCHANGED)
	 return;
 if(value < 0) 
	{
	 fprintf(out_fp, "\n# Warning: input error on Minimum Transient ");
	 fprintf(out_fp, "Size selection!"); 	
         fprintf(out_fp, "\n#          Minimum Transient Size ");
	 fprintf(out_fp, "is now %ld.\n", min_transient_length);
	 return;
	}
 min_transient_length = value;
} 


/*
** Function: void set_tr_bat_size(long value)
** Return  : None
** Remarks : this function may be called to set 'tr_bat_size' value.
**
**	     'value' is the new 'tr_bat_size' value.
**	     Legal range of 'value' : value > 0 .
**
**           Default value of 'value' :  30 (DFLT_TRBTSZ)
**
**           IF 	(value = LONG_UNCHANGED)
**	     THEN 	the default value is assigned to 'tr_bat_size'
**	     ENDIF
** 	
**           IF         (value is out of range)
**           THEN       an error message appears into the proper file and 
**                      the default value is assigned to 'tr_bat_size'
**           ELSE	'value' is assigned to 'tr_bat_size'
**	     ENDIF
**
**
**	     WARNING !! - The call to this function must mandatorily 
**			- follow the call to 'init_statistics(...)'.
**	     
*/
void set_tr_bat_size(long value)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(tr_bat_size != DFLT_TRBTSZ)
	{
	 printf("\nFatal Error on using 'set_tr_bat_size(...)' !!\n");
         printf("\nTransient Batch Size setting is not permitted ");
         printf("before calling \n\n\t'init_statistics(...)' !!\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

 if(value == LONG_UNCHANGED)
	 return;
 if(value <= 0) 
	{
	 fprintf(out_fp, "\n# Warning: input error on Transient Batch ");
	 fprintf(out_fp, "Size selection!"); 	
         fprintf(out_fp, "\n#          Transient Batch Size ");
	 fprintf(out_fp, "is now %ld.\n", tr_bat_size);
	 return;
	}
 tr_bat_size = value;
} 


/*
** Function: void set_tr_seq_len(long value)
** Return  : None
** Remarks : this function may be called to set 'tr_seq_len' value.
**
**	     'value' is the new 'tr_seq_len' value.
**	     Legal range of 'value' : value >= 2 .
**
**           Default value of 'value' :  3 (DFLT_TRSQLN)
**
**
**	     WARNING !!
**
**	     IF 	(value = 0)
**	     THEN	the initial transient detection is not carried out.
**	     ENDIF
**
**
**           IF 	(value = LONG_UNCHANGED)
**	     THEN 	the default value is assigned to 'tr_seq_len'
**	     ENDIF
** 	
**           IF         (value is out of range)
**           THEN       an error message appears into the proper file and 
**                      the default value is assigned to 'tr_seq_len'
**           ELSE	'value' is assigned to 'tr_seq_len'
**	     ENDIF
**
**
**	     WARNING !! - The call to this function must mandatorily 
**			- follow the call to 'init_statistics(...)'.
**	     
*/
void set_tr_seq_len(long value)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(tr_seq_len != DFLT_TRSQLN)
	{
	 printf("\nFatal Error on using 'set_tr_seq_len(...)' !!\n");
         printf("\nTransient Sequence Length setting is not permitted ");
         printf("before calling \n\n\t'init_statistics(...)' !!\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

 if(value == LONG_UNCHANGED)
	 return;
 if(value < 2) 
    {
     if(value == 0)
	{
	 int ii;
	 for(ii = 0; ii < no_stat_var; ii++)	
		 tr_end_flag[ii] = TRUE;
	 fprintf(out_fp, "\n# Warning: the initial transient detection is ");
	 fprintf(out_fp, "not carried out !!\n");
	}
     else
	{
	 fprintf(out_fp, "\n# Warning: input error on Transient Batch Means ");
	 fprintf(out_fp, "Sequence Length selection!"); 	
         fprintf(out_fp, "\n#          Transient Batch Means Sequence Length ");
	 fprintf(out_fp, "is now %ld.\n", tr_seq_len);
	 return;
	}
     }
 if(value * no_stat_var > STAT_ARR_SIZE)
	{
	 tr_seq_len = STAT_ARR_SIZE / no_stat_var;

	 fprintf(out_fp, "\n# Warning: input error on Transient Batch Means ");
	 fprintf(out_fp, "Sequence Length selection!"); 	
         fprintf(out_fp, "\n#          Transient Batch Means Sequence Length ");
	 fprintf(out_fp, "is now %ld.\n", tr_seq_len);
	 return;
	}
 tr_seq_len = value;
} 


/*
** Function: void set_tr_accuracy(double value)
** Return  : None
** Remarks : this function may be called to set 'tr_accuracy' value.
**
**	     'value' is the new 'tr_accuracy' value.
**	     Legal range of 'value' : (0.0, 1.0) .
**
**           Default value of 'tr_accuracy' :  0.99 (DFLT_TRACC)
**
**           IF 	(value = DOUBLE_UNCHANGED)
**	     THEN 	the default value is assigned to 'tr_accuracy'
**	     ENDIF
** 	
**           IF         (value is out of range)
**           THEN       an error message appears into the proper file and 
**                      the default value is assigned to 'tr_accuracy'
**           ELSE	'value' is assigned to 'tr_accuracy'
**	     ENDIF
**
**
**	     WARNING !! - The call to this function must mandatorily 
**			- follow the call to 'init_statistics(...)'.
**	    
**	     
*/
void set_tr_accuracy(double value)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(tr_accuracy != DFLT_TRACC)
	{
	 printf("\nFatal Error on using 'set_tr_accuracy(...)' !!\n");
         printf("\nTransient Accuracy level setting is not permitted ");
         printf("before calling \n\n\t'init_statistics(...)' !!\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

 if(value == DOUBLE_UNCHANGED)
	 return;
 if((value <= 0.0) || (value >= 1.0))
	{
	 fprintf(out_fp, "\n# Warning: your Transient Accuracy was out ");
	 fprintf(out_fp, "of range!");	
	 fprintf(out_fp, "\n#          Accuracy is now %f.\n", tr_accuracy);	
	 return;
	}
 tr_accuracy = value;
} 


/*
** Function: void set_st_bat_size(long value)
** Return  : None
** Remarks : this function may be called to set 'st_bat_size' value.
**
**	     'value' is the new 'st_bat_size' value.
**	     Legal range of 'value' : value >= 2 .
**
**           Default value of 'value' :  10000 (DFLT_STBTSZ)
**
**           IF 	(value = LONG_UNCHANGED)
**	     THEN 	the default value is assigned to 
**                      'st_bat_size'
**	     ENDIF
** 	
**           IF         (value is out of range)
**           THEN       an error message appears into the proper file and 
**                      the default value is assigned to 'st_bat_size'
**           ELSE	'value' is assigned to 'st_bat_size'
**	     ENDIF
**
**
**	     WARNING !! - The call to this function must mandatorily 
**			- follow the call to 'init_statistics(...)'.
**	     
*/
void set_st_bat_size(long value)
{
 static Booltype function_already_called = FALSE;

 if(function_already_called == FALSE)
   {
    if(st_bat_size != DFLT_STBTSZ)
	{
	 printf("\nFatal Error on using 'set_st_bat_size(...)' !!\n");
         printf("\nStationary Batch Size setting is not permitted ");
         printf("before calling \n\n\t'init_statistics(...)' !!\n");
	 exit(1);
	}	 
    function_already_called = TRUE;
   }

 if(value == LONG_UNCHANGED)
	 return;
 if(value < 2) 
	{
	 fprintf(out_fp, "\n# Warning: input error on Stationary Batch ");
	 fprintf(out_fp, "Size selection!"); 	
         fprintf(out_fp, "\n#          Stationary Batch Size ");
	 fprintf(out_fp, "is now %ld.\n", st_bat_size);
	 return;
	}
 st_bat_size = value;
} 

